package compilo;
import static org.junit.Assert.*;

import java.io.IOException;
import java.io.InputStream;

import org.antlr.runtime.ANTLRInputStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.Tree;
import org.junit.Test;

public class ExprGenTest {

	private String exec(String test) {
		final char[] strInput = test.toCharArray();
		ANTLRInputStream input;
		Tree tree = null;

		try {
			input = new ANTLRInputStream(new InputStream() {
				private int i = 0;
				
				@Override
				public int read() throws IOException {
					if(i - strInput.length < 0)
						return strInput[i++];
					
					if(i - strInput.length == 0) {
						i++;
						return '\n';
					}
					
					return -1;
				}
			});

			ExprLexer lexer = new ExprLexer(input);
			ExprParser parser = new ExprParser(new CommonTokenStream(lexer));
			tree = (Tree) parser.prog().getTree();
			
			if(tree.toString().length() > 2 && tree.toString().startsWith("<")) {
				System.err.println("Grammar error!");
				fail();
			}

		} catch (IOException e) {
			e.printStackTrace();
		} catch (RecognitionException e) {
			e.printStackTrace();
		}

		ExprSemGen tse = new ExprSemGen(true);

		try {
			tse.complete(tree);
		} catch (SemException e) {
			e.printError();
		}
		
		return tse.getOut();
	}

	@Test
	public void varTest() {
		assertEquals("int toto", exec("var toto : int"));
		assertEquals("int toto=8", exec("var toto : int := 8"));
		assertEquals("int toto=8", exec("var toto := 8"));
		assertEquals("int toto[42]", exec("var toto : int[42]"));
		assertEquals("int toto[3]={1,2,3}", exec("var toto : int[3] := { 1, 2, 3 }"));
	}
	
	@Test
	public void funCallTest() {
		assertEquals(cfToto() + "toto();", exec(fToto() + "toto()"));

		assertEquals(
					"int titi(int c,int d){1;};" +
					"int a=1;" +
					"titi(a,42);",
				exec(
					"function titi(c : int, d : int) : int = 1\n" +
					"var a := 1\n" +
					"titi(a, 42)")
				);
	}
	
	@Test
	public void funDecTest() {
		assertEquals("int toto(){1;}", exec("function toto() : int = 1"));
		assertEquals("int toto(int a,char* b){1;}", exec("function toto(a : int, b : string) : int = 1"));
		
		// /!\ <  void toto(){}  > but was <  void toto(){;}  >
		assertEquals("void toto(){}", exec("function toto() = ()"));
		
		
	}
	
	@Test
	public void idOrAffTest() {
		assertEquals("int a=1;a;", exec("var a := 1\na"));
		assertEquals("int a;a=1;", exec("var a : int\na := 1"));
		assertEquals(cfToto() + "int a;a=toto();", exec(fToto() + "var a : int\na := toto()"));
		assertEquals("int a[2]={1,2};a[1];", exec("var a : int[2] := {1,2}\na[1]"));
		assertEquals(cfToto() + "int a[2];a[toto()];", exec(fToto() + "var a : int[2]\na[toto()]"));
		assertEquals("int a[2];a[1]=1;", exec("var a : int[2]\na[1] := 1"));
	}
	
	@Test
	public void seqTest() {
		assertEquals("int a=1;(a);", exec("var a := 1\n(a)"));
		assertEquals("", exec("()"));
		assertEquals(cfToto() + "int a=1;(a,toto(),1);", exec(fToto() + "var a := 1\n(a; toto(); 1)"));
	}
	
	@Test
	public void blockTest() {
		assertEquals("{}", exec("let in end"));
		assertEquals("{int a=1;}", exec("let var a := 1\n in end"));
		assertEquals("{1;}", exec("let in 1\n end"));
		assertEquals(cfToto() + "{int a;a=toto();};", exec(fToto() + "let var a : int\n in a := toto()\n end"));
	}
	
	@Test
	public void forTest() {
		assertEquals("int i=0;for(;i<=10;i++)1", exec("for i := 0 to 10 do 1"));
		assertEquals("int i=0;for(;i<=10;i++){1;2;3;}", exec("for i := 0 to 10 do let in 1\n2\n3\n end"));
	}
	
	@Test
	public void whileTest() {
		assertEquals("while(1)1", exec("while 1 do 1"));
		assertEquals(cfToto() + "while(toto()){1;toto();};", exec(fToto() + "while toto() do let in 1\ntoto()\n end"));
	}
	
	@Test
	public void ifTest() {
		assertEquals("if(1)1", exec("if 1 then 1"));
		assertEquals("if(1)1; else 2", exec("if 1 then 1 else 2"));
	}
	
	@Test
	public void opTest() {
		assertEquals("1<2", exec("1<2"));
		assertEquals("1==2", exec("1=2"));
		assertEquals("if(2>1)3+1", exec("if 2 > 1 then 3 + 1"));
	}
	
	
	/* OUTILS */
	
	private String cfToto() {
		return "int toto(){1;};";
	}
	
	private String fToto() {
		return "function toto() : int = 1\n";
	}
}
