package compilo;
import static org.junit.Assert.*;

import java.io.IOException;
import java.io.InputStream;
import org.antlr.runtime.ANTLRInputStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.Tree;
import org.junit.Test;

public class ExprSemTest {
	
	private TSElement.Type exec(String test) throws SemException {
		final char[] strInput = test.toCharArray();
		ANTLRInputStream input;
		Tree tree = null;

		try {
			input = new ANTLRInputStream(new InputStream() {
				private int i = 0;
				
				@Override
				public int read() throws IOException {
					if(i - strInput.length < 0)
						return strInput[i++];
					
					if(i - strInput.length == 0) {
						i++;
						return '\n';
					}
					
					return -1;
				}
			});

			ExprLexer lexer = new ExprLexer(input);
			ExprParser parser = new ExprParser(new CommonTokenStream(lexer));
			tree = (Tree) parser.prog().getTree();
			
			if(tree.toString().startsWith("<")) {
				System.err.println("Grammar error!");
				fail();
			}

		} catch (IOException e) {
			e.printStackTrace();
		} catch (RecognitionException e) {
			e.printStackTrace();
		}

		return new ExprSemGen().complete(tree);
	}
	
	@Test
	public void varWrongValueTypeTest() {
		try {
			exec("var toto : int := \"lala\"");
			fail();
			
		} catch (SemException e) {
			assertEquals("Wrong value type (STRING value with INT var)", e.getMessage());
		}
	}
	
	@Test
	public void varExtractTypeFromValueTest() {
		TSElement.Type type = null;
		
		try {
			type = exec("var toto := 3");
			assertEquals(TSElement.Type.INT, type);
			
		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void varDecTest() {
		try {
			exec("var toto : string");
			assertTrue(true);
			
		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void varTabAffTest() {
		try {
			exec("var toto : int[2] := { 1, 2 }");
			assertTrue(true);
			
		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void varTabOutOfBoundsTest() {
		try {
			exec(
				"var toto : int[2] := { 1, 2 }\n" +
				"toto[2]"
			);
			fail();
			
		} catch (SemException e) {
			assertEquals("2: index out of bounds", e.getMessage());
		}
	}
	
	@Test
	public void varTabSizeTypeTest() {
		try {
			exec(
				"var toto : int[10]\n" +
				"var t := 2\n" +
				"toto[t]"
			);
			assertTrue(true);
			
		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void varTabWrongValueTest() {
		try {
			exec("var toto : string[2] := { 1, 2 }");
			fail();
			
		} catch (SemException e) {
			assertEquals("Wrong tab value type (INT value with STRING tab)", e.getMessage());
		}
	}
	
	@Test
	public void varTabWrongNumberTest() {
		try {
			exec("var toto : int[3] := { 1, 2 }");
			fail();
			
		} catch (SemException e) {
			assertEquals("Wrong number of values for this tab (2 values in a 3-tab)", e.getMessage());
		}
	}
	
	@Test
	public void varDuplicate() {
		try {
			exec(
				"var toto : int\n" + 
				"var toto : string[2]"
			);
			fail();
			
		} catch (SemException e) {
			assertEquals("A INT named toto already exists", e.getMessage());
		}
	}
	
	@Test
	public void funCallNotfoundTest() {
		try {
			exec("toto()");
			fail();
			
		} catch (SemException e) {
			assertEquals("The called function toto doesn't exist", e.getMessage());
		}
	}
	
	@Test
	public void funDecCallWrongNumberTest() {
		try {
			exec(
				"function test1(ch: string) : int = 10\n" + 
				"test1(42)"
			);
			fail();
			
		} catch (SemException e) {
			assertEquals("test1: wrong arg type (INT instead of STRING)", e.getMessage());
		}
	}
	
	@Test
	public void funDecWrongReturnTest() {
		try {
			exec("function test1(ch: string) : int = 4.2");
			fail();
			
		} catch (SemException e) {
			assertEquals("test1 must return a INT instead of a REAL", e.getMessage());
		}
	}
	
	@Test
	public void funDecWithNoArgTest() {
		try {
			exec("function test1() = 10");
			assertTrue(true);
			
		} catch (SemException e) {
			e.printError();
			fail();
		}
	}

	@Test
	public void varDecSeqRecTest() {
		try {
			exec("var toto : real := (42; 41; (40; 39; (38; 3.7)))");
			assertTrue(true);
			
		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void varDecSeqWrongTypeTest() {
		try {
			exec("var toto : int := (42; 4.2)");
			fail();
			
		} catch (SemException e) {
			assertEquals("Wrong value type (REAL value with INT var)", e.getMessage());
		}
	}
	
	@Test
	public void funDecInnerDuplicateTest() {
		try {
			exec("function test1(ch: string) : int = ( var ch := 4.2 )");
			fail();
			
		} catch (SemException e) {
			assertEquals("A STRING named ch already exists", e.getMessage());
		}
	}
	
	@Test
	public void idOrAffTest() {
		try {
			exec(
				"var a : int\n" + 
				"var b := 42\n" +
				"a := b"
			);
			assertTrue(true);

		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void idOrAffWrongTypeTest() {
		try {
			exec(
				"var a : int\n" +
				"var b := 4.2\n" +
				"a := b"
			);
			fail();

		} catch (SemException e) {
			assertEquals("Wrong value type (REAL value with INT var)", e.getMessage());
		}
	}
	
	@Test
	public void idOrAffTabWrongTypeTest() {
		try {
			exec(
				"var toto : int[42]\n" +
				"toto[12] := 4.2"
			);
			fail();

		} catch (SemException e) {
			assertEquals("Wrong tab value type (REAL value with INT tab)", e.getMessage());
		}
	}

	@Test
	public void idOrAffTabNotTabTest() {
		try {
			exec(
				"var toto : int\n" +
				"toto[2] := 42"
			);
			fail();

		} catch (SemException e) {
			assertEquals("toto is not a tab", e.getMessage());
		}
	}
	
	@Test
	public void idOrAffTabNotVarTest() {
		try {
			exec(
				"var toto : int[42]\n" +
				"toto := 21"
			);
			fail();

		} catch (SemException e) {
			assertEquals("toto is a tab, you have to use an index", e.getMessage());
		}
	}
	
	@Test
	public void idOrAffTabWrongFunTypeTest() {
		try {
			exec(
				"function a() : real = 4.2\n" +
				"var toto : int[42]\n" +
				"toto[a()] := 21"
			);
			fail();

		} catch (SemException e) {
			assertEquals("FUN_CALL must return or be a INT (not a REAL)", e.getMessage());
		}
	}
	
	@Test
	public void idNotAffAndMiscTest() {
		try {
			exec(
				"function b() : int = 1\n" +
				"var toto : int[42]\n" +
				"var titi : int[2]\n" +
				"titi[1] := 4\n" +
				"toto[b()]\n" +
				"titi[1]\n" +
				"toto[titi[1]]\n" +
				"toto[20]"
			);
			fail();

		} catch (SemException e) {
			assertEquals("The value at the index 20 was not set", e.getMessage());
		}
	}

	@Test
	public void blockEmptyTest() {
		try {
			exec("let in end");
			assertTrue(true);

		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void blockTest() {
		try {
			exec(
				"let\n" + 
				"	var toto := 3\n" +
				"in\n" +
				"	toto := 3\n" +
				"end"
			);
			assertTrue(true);

		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void forIdAffTest() {
		try {
			exec("for i := 0 to 21 do 42");
			assertTrue(true);

		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void forROTest() {
		try {
			exec("for i := 0 to 21 do ( i := 10 )");
			fail();

		} catch (SemException e) {
			assertEquals("i is a read-only var in this context", e.getMessage());
		}
	}
	
	@Test
	public void forParamFunCallTest() {
		try {
			exec(
				"function max() : int = 10\n" +
				"for i := max() to max() do ( i )"
			);
			assertTrue(true);

		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void forWhileTest() {
		try {
			exec("while 1 do ( 1; 2 )");
			assertTrue(true);

		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void forIfTest() {
		try {
			exec("if 1 then 2");
			assertTrue(true);

		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	// pas de retour à la ligne avant le else :( !
	@Test
	public void forIfElseTest() {
		try {
			exec("if 1 then 2 else 3");
			assertTrue(true);

		} catch (SemException e) {
			e.printError();
			fail();
		}
	}
	
	@Test
	public void forIfElseVarTest() {
		try {
			exec("if 1 then var i : int else ( i )");
			assertTrue(true);

		} catch (SemException e) {
			assertEquals("i doesn't exist", e.getMessage());
		}
	}
	
	@Test
	public void opTypeTest() {
		try {
			exec("var toto : int := 1 + 2.3");
			fail();

		} catch (SemException e) {
			assertEquals("Wrong value type (REAL value with INT var)", e.getMessage());
		}
	}
	
	@Test
	public void opTypeStringTest() {
		try {
			exec(
				"function i() : int = 1\n" +
				"var toto : int := i() + \"toto\""
			);
			fail();

		} catch (SemException e) {
			assertEquals("\"toto\": arithmetic op needs numbers only", e.getMessage());
		}
	}
}